"""
Distance metrics
"""

from typing import Callable
import numpy as np
from pyfit import Tensor

# Distance function type
Distance = Callable[[Tensor, Tensor], float]


def euclidean_distance(a: Tensor, b: Tensor) -> float:
    """
    Euclidean distance: https://en.wikipedia.org/wiki/Euclidean_distance
    """
    squared_diff: Tensor = (a - b) ** 2
    sum_squared_diff: float = np.sum(squared_diff)
    eucl_dist: float = np.sqrt(sum_squared_diff)
    return eucl_dist



"""
Utilities for preparing data before training
"""

from typing import List
import numpy as np
from pyfit import Tensor


def train_test_split(*x: Tensor, test_ratio: float = 0.25) -> List[Tensor]:
    """
    Split and shuffle one or several same-length tensor(s) between training and test sets,
    according to a chosen test ratio
    """
    n_samples: int = x[0].shape[0]
    split_index: int = round(n_samples * (1 - test_ratio))
    # https://stackoverflow.com/a/4602224
    permutation: Tensor = np.random.permutation(n_samples)
    splitted = []
    for tensor in x:
        shuffled_tensor = tensor[permutation]
        if shuffled_tensor.ndim > 1:
            splitted.append(shuffled_tensor[:split_index, :])
            splitted.append(shuffled_tensor[split_index:, :])
        else:
            splitted.append(shuffled_tensor[:split_index])
            splitted.append(shuffled_tensor[split_index:])
    return splitted


def scale_min_max(x: Tensor) -> Tensor:
    """
    Scale a tensor into the [0..1] range
    """
    # Compute min and max feature-wise
    min_x: Tensor = x.min(axis=0)
    max_x: Tensor = x.max(axis=0)
    return (x - min_x) / (max_x - min_x)


def scale_standard(x: Tensor, *, mean: Tensor = None, std: Tensor = None) -> Tensor:
    """
    Scale a tensor to zero mean and unit variance
    """
    if mean is None:
        # Compute mean feature-wise
        mean = x.mean(axis=0)
    if std is None:
        # Compute standard deviation feature-wise
        std = x.std(axis=0)
    return (x - mean) / std


def vectorize_sequences(sequences: Tensor, dimension: int = 10000) -> Tensor:
    """One-hot encode a vector of sequences into a binary matrix (number of sequences, dimension)"""

    # Example : [[3, 5]] -> [[0. 0. 0. 1. 0. 1. 0...]]

    results = np.zeros((len(sequences), dimension))
    # set specific indices of results[i] to 1s
    for i, sequence in enumerate(sequences):
        results[i, sequence] = 1.0
    return results



"""
Regression metrics
"""

import numpy as np
from pyfit import Tensor


def mean_squared_error(y_true: Tensor, y_pred: Tensor) -> float:
    """
    Compute Mean Squared Error
    """
    squared_error: Tensor = (y_true - y_pred) ** 2
    mse: float = np.mean(squared_error)
    return mse



"""
Model classes
"""

from typing import Sequence
from pyfit import Tensor
from ._layers import Differentiable


class Sequential(Differentiable):
    """
    Neural network defined as a linear stack of layers
    """

    def __init__(self, layers: Sequence[Differentiable]) -> None:
        self.layers = layers

    def forward(self, inputs: Tensor) -> Tensor:
        for layer in self.layers:
            inputs = layer.forward(inputs)
        return inputs

    def backward(self, gradients: Tensor) -> Tensor:
        for layer in reversed(self.layers):
            gradients = layer.backward(gradients)
        return gradients



"""
Unit tests for preprocessing functions
"""

# Docstrings are superfluous for test functions
# pylint: disable=missing-docstring

from math import isclose
import numpy as np
from pyfit.preprocessing import train_test_split, scale_min_max, scale_standard


def test_train_test_split() -> None:
    # Generate a random 1d tensor
    x = np.random.randint(1, 10, (17,))

    x_train, x_test = train_test_split(x, test_ratio=0.3)
    assert x_train.shape == (12,)
    assert x_test.shape == (5,)

    # Generate a random 2d tensor
    x = np.random.rand(30, 3)

    x_train, x_test = train_test_split(x, test_ratio=0.33)
    assert x_train.shape == (20, 3)
    assert x_test.shape == (10, 3)

    # Default ratio is 0.25
    x_train, x_test = train_test_split(x)
    assert x_train.shape == (22, 3)
    assert x_test.shape == (8, 3)


def test_train_test_split_2tensors() -> None:
    # Generate two tensors with same length (first dimension)
    x = np.random.rand(30, 3)
    y = np.random.rand(30)

    x_train, x_test, y_train, y_test = train_test_split(x, y)

    assert x_train.shape == (22, 3)
    assert x_test.shape == (8, 3)
    assert y_train.shape == (22,)
    assert y_test.shape == (8,)


def test_scale_min_max() -> None:
    # Generate a random (3,3) tensor with values between 1 and 10
    x = np.random.randint(1, 10, (3, 3))

    x_scaled = scale_min_max(x)
    assert x_scaled.min() == 0
    assert x_scaled.max() == 1


def test_scale_standard() -> None:
    # Generate a random tensor with int values between 1 and 10
    x = np.random.randint(1, 10, (3, 3))

    x_scaled = scale_standard(x)

    mean_scaled: float = x_scaled.mean()
    std_scaled: float = x_scaled.std()
    # https://stackoverflow.com/a/35325039
    assert isclose(mean_scaled, 0, abs_tol=1.0e-8)
    assert isclose(std_scaled, 1)


def test_scale_standard_mean_std() -> None:
    # Generate a random tensor with int values between 1 and 10
    x = np.random.randint(1, 10, (3, 4))

    # Scale with zero mean and unit variance (should do nothing)
    zero_mean = np.zeros(x.shape[1])
    unit_variance = np.ones(x.shape[1])
    x_scaled = scale_standard(x, mean=zero_mean, std=unit_variance)

    # Check that mean and standard deviation haven't been modified by scaling
    assert x_scaled.mean() == x.mean()
    assert x_scaled.std() == x.std()



"""
Core definitions for models and training
"""

# pylint: disable=unused-import

# Temporarily disable annoying Pylint check
# pylint: disable=too-few-public-methods

import numpy as np

# A tensor is a multidimensional array
# Use NumPy n-dimensional array class as our Tensor class
from numpy import ndarray as Tensor


class BaseEstimator:
    """
    Abstract base class for all estimators
    """

    x_train: Tensor = None
    y_train: Tensor = None

    def fit(self, x_train: Tensor, y_train: Tensor) -> None:
        """
        Fit the model using x_train as training data and y_train as target values
        """
        # Stores training data and targets for use in derived classes
        self.x_train = x_train
        self.y_train = y_train



"""
Activation functions
"""

# Docstrings are superfluous for activation functions
# pylint: disable=missing-docstring

import numpy as np
from pyfit import Tensor


def tanh(x: Tensor) -> Tensor:
    return np.tanh(x)


def tanh_prime(x: Tensor) -> Tensor:
    y: Tensor = tanh(x)
    return 1 - y ** 2



"""
Data generation functions
"""

# Allow uppercase and non-snake_case names
# pylint: disable=invalid-name

from typing import Tuple
import numpy as np
from pyfit import Tensor


def make_multiclass(
    n_points: int = 500, n_dim: int = 2, n_classes: int = 3
) -> Tuple[Tensor, Tensor]:
    """
    Generate spiral-shaped data
    n: number of points per class
    d: dimensionality
    k: number of classes
    """

    np.random.seed(0)
    x = np.zeros((n_points * n_classes, n_dim))
    y = np.zeros(n_points * n_classes)
    for j in range(n_classes):
        ix = range(n_points * j, n_points * (j + 1))
        # radius
        r = np.linspace(0.0, 1, n_points)
        # theta
        t = np.linspace(j * 4, (j + 1) * 4, n_points) + np.random.randn(n_points) * 0.2
        x[ix] = np.c_[r * np.sin(t), r * np.cos(t)]
        y[ix] = j
    return x, y



"""
Classification metrics
"""

import numpy as np
from pyfit import Tensor


def accuracy(y_true: Tensor, y_pred: Tensor) -> float:
    """
    Compute accuracy: number of exact predictions / total number of predictions
    """
    n_exact: int = np.sum(y_pred == y_true)
    n_total: int = y_true.size
    return n_exact / n_total



"""
Unit tests for neighbors-based algorithms
"""

# Docstrings are superfluous for test functions
# pylint: disable=missing-docstring

import numpy as np
from pyfit.neighbors import KNeighborsClassifier, KNeighborsRegressor


def test_knn_classifier_int() -> None:
    samples = np.array([[0], [1], [2], [3]])
    targets = np.array([0, 0, 1, 1])

    clf = KNeighborsClassifier()
    clf.fit(samples, targets)

    predicted = clf.predict(np.array([[1.1]]))
    assert predicted[0] == 0

    predicted = clf.predict(np.array([[2.2], [0.9]]))
    assert predicted[0] == 1
    assert predicted[1] == 0


def test_knn_classifier_str() -> None:
    samples = np.array([[0], [1], [2], [3]])
    targets = np.array(["a", "a", "b", "b"])

    clf = KNeighborsClassifier()
    clf.fit(samples, targets)

    predicted = clf.predict(np.array([[1.1]]))
    assert predicted[0] == "a"

    predicted = clf.predict(np.array([[2.2], [0.9]]))
    assert predicted[0] == "b"
    assert predicted[1] == "a"


def test_knn_regressor() -> None:
    samples = np.array([[0], [1], [2], [3]])
    targets = np.array([0, 0, 1, 1])

    reg = KNeighborsRegressor(k=2)
    reg.fit(samples, targets)

    predicted = reg.predict(np.array([[1.5]]))
    assert predicted[0] == 0.5



"""
Plotting functions
"""

# Allow uppercase names
# pylint: disable=invalid-name

from typing import Any
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from pyfit import Tensor


def plot_planar_data(X: Tensor, y: Tensor) -> None:
    """Plot some planar data"""

    plt.figure()
    plt.plot(X[y == 0, 0], X[y == 0, 1], "or", alpha=0.5, label=0)
    plt.plot(X[y == 1, 0], X[y == 1, 1], "ob", alpha=0.5, label=1)
    plt.legend()


def plot_decision_boundary(
    pred_func: Any, X: Tensor, y: Tensor, figure: Any = None
) -> None:
    """Plot a decision boundary"""

    if figure is None:  # If no figure is given, create a new one
        plt.figure()
    # Set min and max values and give it some padding
    x_min, x_max = X[:, 0].min() - 0.5, X[:, 0].max() + 0.5
    y_min, y_max = X[:, 1].min() - 0.5, X[:, 1].max() + 0.5
    h = 0.01
    # Generate a grid of points with distance h between them
    xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
    # Predict the function value for the whole grid
    Z = pred_func(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)
    # Plot the contour and training examples
    plt.contourf(xx, yy, Z, cmap=plt.cm.get_cmap("Spectral"))
    cm_bright = ListedColormap(["#FF0000", "#0000FF"])
    plt.scatter(X[:, 0], X[:, 1], c=y, cmap=cm_bright)


def plot_multiclass(X: Tensor, y: Tensor) -> None:
    """Plot spiral-shaped data"""

    plt.figure()
    plt.scatter(X[:, 0], X[:, 1], c=y, s=40, cmap=plt.cm.get_cmap("Set1"), alpha=0.8)
    plt.xlim([-1, 1])
    plt.ylim([-1, 1])


def plot_multiclass_decision_boundary(pred_func: Any, X: Tensor, y: Tensor) -> None:
    """Plot a multiclass decision boundary"""

    x_min, x_max = X[:, 0].min() - 0.1, X[:, 0].max() + 0.1
    y_min, y_max = X[:, 1].min() - 0.1, X[:, 1].max() + 0.1
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 101), np.linspace(y_min, y_max, 101))
    Z = pred_func(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)
    plt.figure()
    plt.contourf(xx, yy, Z, cmap=plt.cm.get_cmap("tab20b_r"))
    plt.scatter(X[:, 0], X[:, 1], c=y, s=40, cmap=plt.cm.get_cmap("Set1"))
    plt.xlim(xx.min(), xx.max())
    plt.ylim(yy.min(), yy.max())


def plot_loss_acc(history: Any) -> None:
    """Plot training and (optionally) validation loss and accuracy"""

    loss = history.history["loss"]
    epochs = range(1, len(loss) + 1)

    plt.figure()

    plt.subplot(2, 1, 1)
    plt.plot(epochs, loss, ".--", label="Training loss")
    final_loss = loss[-1]
    title = "Training loss: {:.4f}".format(final_loss)
    plt.ylabel("Loss")
    if "val_loss" in history.history:
        val_loss = history.history["val_loss"]
        plt.plot(epochs, val_loss, "o-", label="Validation loss")
        final_val_loss = val_loss[-1]
        title += ", Validation loss: {:.4f}".format(final_val_loss)
    plt.title(title)
    plt.legend()

    acc = history.history["acc"]

    plt.subplot(2, 1, 2)
    plt.plot(epochs, acc, ".--", label="Training acc")
    final_acc = acc[-1]
    title = "Training accuracy: {:.2f}%".format(final_acc * 100)
    plt.xlabel("Epochs")
    plt.ylabel("Accuracy")
    if "val_acc" in history.history:
        val_acc = history.history["val_acc"]
        plt.plot(epochs, val_acc, "o-", label="Validation acc")
        final_val_acc = val_acc[-1]
        title += ", Validation accuracy: {:.2f}%".format(final_val_acc * 100)
    plt.title(title)
    plt.legend()


def plot_loss_mae(history: Any) -> None:
    """Plot training and (optionally) validation loss and MAE"""

    loss = history.history["loss"]
    epochs = range(1, len(loss) + 1)

    plt.figure()

    plt.subplot(2, 1, 1)
    plt.plot(epochs, loss, ".--", label="Training loss")
    final_loss = loss[-1]
    title = "Training loss: {:.4f}".format(final_loss)
    plt.ylabel("Loss")
    if "val_loss" in history.history:
        val_loss = history.history["val_loss"]
        plt.plot(epochs, val_loss, "o-", label="Validation loss")
        final_val_loss = val_loss[-1]
        title += ", Validation loss: {:.4f}".format(final_val_loss)
    plt.title(title)
    plt.legend()

    mae = history.history["mean_absolute_error"]

    plt.subplot(2, 1, 2)
    plt.plot(epochs, mae, ".--", label="Training MAE")
    final_mae = mae[-1]
    title = "Training MAE: {:.2f}".format(final_mae)
    plt.xlabel("Epochs")
    plt.ylabel("MAE")
    if "val_mean_absolute_error" in history.history:
        val_mae = history.history["val_mean_absolute_error"]
        plt.plot(epochs, val_mae, "o-", label="Validation MAE")
        final_val_mae = val_mae[-1]
        title += ", Validation MAE: {:.2f}".format(final_val_mae)
    plt.title(title)
    plt.legend()



"""
K-Nearest Neighbors algorithms
"""

# Temporarily disable annoying Pylint check
# pylint: disable=too-few-public-methods

from collections import Counter
from typing import List, Union, Tuple
import numpy as np
from pyfit import Tensor, BaseEstimator
from pyfit.metrics import Distance, euclidean_distance


class KNeighborsEstimator(BaseEstimator):
    """
    Abstract K-NN estimator
    """

    def __init__(self, *, k: int = 5, distance: Distance = euclidean_distance) -> None:
        super().__init__()
        self.k = k
        self.distance = distance

    def _sort_neighbors(self, x_new: Tensor) -> List[Tuple[Tensor, Tensor]]:
        """
        Sort all samples by (increasing) distance to new sample
        """
        # Associate training data and targets in a list of tuples
        training_samples: List[Tuple[Tensor, Tensor]] = list(
            zip(self.x_train, self.y_train)
        )

        # Sort samples by their distance to new sample
        sorted_neighbors: List[Tuple[Tensor, Tensor]] = sorted(
            ((x, y) for (x, y) in training_samples),
            key=lambda sample: self.distance(x_new, sample[0]),
        )

        return sorted_neighbors


class KNeighborsClassifier(KNeighborsEstimator):
    """
    K-NN classifier
    """

    def predict(self, x_test: Tensor) -> List[Union[int, str]]:
        """
        Predict results for test samples
        """
        # Return a list containing the prediction for each new sample
        return [self._predict_one(x_new) for x_new in x_test]

    def _predict_one(self, x_new: Tensor) -> Union[int, str]:
        """
        Predict result for a single new sample
        """
        neighbors = self._sort_neighbors(x_new)

        # Keep targets of the k nearest neighbors
        k_nearest_targets: List[Union[int, str]] = [
            target for (_, target) in neighbors[: self.k]
        ]

        # Get most frequent target in nearest neighbors
        winner_target: Union[int, str] = Counter(k_nearest_targets).most_common(1)[0][0]

        return winner_target


class KNeighborsRegressor(KNeighborsEstimator):
    """
    K-NN regressor
    """

    def predict(self, x_test: Tensor) -> List[float]:
        """
        Predict results for test samples
        """
        # Return a list containing the prediction for each new sample
        return [self._predict_one(x_new) for x_new in x_test]

    def _predict_one(self, x_new: Tensor) -> float:
        """
        Predict result for a single new sample
        """
        neighbors = self._sort_neighbors(x_new)

        # Keep targets of the k nearest neighbors
        k_nearest_targets: List[float] = [target for (_, target) in neighbors[: self.k]]

        # Compute the mean target value
        target_mean = np.mean(k_nearest_targets)

        return target_mean



"""
Unit tests for metrics
"""

# Docstrings are superfluous for test functions
# pylint: disable=missing-docstring

from math import isclose
import numpy as np
from pyfit.metrics import mean_squared_error, euclidean_distance, accuracy


def test_mse() -> None:
    y_true = np.array([3, -0.5, 2, 7])
    y_pred = np.array([2.5, 0.0, 2, 8])

    error = mean_squared_error(y_true, y_pred)
    assert isclose(error, (0.25 + 0.25 + 1) / 4)

    y_true = np.array([[1, 0, 0, 1], [0, 1, 1, 1], [1, 1, 0, 1]])
    y_pred = np.array([[0, 0, 0, 1], [1, 0, 1, 1], [0, 0, 0, 1]])

    error = mean_squared_error(y_true, y_pred)
    assert isclose(error, (1.0 / 3 + 2.0 / 3 + 2.0 / 3) / 4)


def test_euclidean_distance() -> None:
    a = np.array([1, 1.5])
    b = np.array([3, 3])

    dist = euclidean_distance(a, b)
    assert isclose(dist, np.linalg.norm(a - b))

    a = np.random.rand(3, 3)
    b = np.random.rand(3, 3)

    dist = euclidean_distance(a, b)
    assert isclose(dist, np.linalg.norm(a - b))


def test_accuracy() -> None:
    y_true = np.array([0, 1, 2, 3])
    y_pred = np.array([0, 2, 1, 3])

    acc = accuracy(y_true, y_pred)
    assert acc == 0.5

    acc = accuracy(y_true, y_true)
    assert acc == 1.0



"""Setup for the pyfit package."""

import setuptools


with open("README.md") as f:
    README = f.read()

setuptools.setup(
    author="Baptiste Pesquet",
    author_email="bpesquet@gmail.com",
    name="pyfit",
    license="MIT",
    description="A simple Machine Learning library",
    version="0.0.3",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/bpesquet/pyfit",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    install_requires=["numpy", "matplotlib"],
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ],
)



"""
Layer classes
"""

# Temporarily disable annoying Pylint check
# pylint: disable=too-few-public-methods

from typing import Dict, Callable
import numpy as np
from pyfit import Tensor
from ._activations import tanh, tanh_prime

F = Callable[[Tensor], Tensor]


class Differentiable:
    """
    Base class for all differentiable computations
    """

    def forward(self, inputs: Tensor) -> Tensor:
        """
        Produce the output corresponding to these inputs
        """
        raise NotImplementedError

    def backward(self, gradients: Tensor) -> Tensor:
        """
        Backpropagate these gradients through the layer
        """
        raise NotImplementedError


class ParametersMixin:
    """
    Mixin for adding parameters and gradient values to a class
    """

    def __init__(self) -> None:
        self.params: Dict[str, Tensor] = {}
        self.grads: Dict[str, Tensor] = {}


class Linear(Differentiable, ParametersMixin):
    """
    Linear layer
    """

    def __init__(self, *, in_features: int, out_features: int) -> None:
        super().__init__()
        # Randomly init weights and bias
        self.params["w"] = np.random.randn(in_features, out_features)
        self.params["b"] = np.random.randn(out_features)
        # Init inputs
        self.inputs: Tensor = None

    def forward(self, inputs: Tensor) -> Tensor:
        self.inputs = inputs
        return inputs @ self.params["w"] + self.params["b"]

    def backward(self, gradients: Tensor) -> Tensor:
        self.grads["b"] = np.sum(gradients, axis=0)
        self.grads["w"] = self.inputs.T @ gradients
        return gradients @ self.params["w"].T


class Activation(Differentiable):
    """
    Layer corresponding to an activation function
    """

    def __init__(self, f: F, f_prime: F) -> None:
        super().__init__()
        self.f = f
        self.f_prime = f_prime
        # Init inputs
        self.inputs: Tensor = None

    def forward(self, inputs: Tensor) -> Tensor:
        self.inputs = inputs
        return self.f(inputs)

    def backward(self, gradients: Tensor) -> Tensor:
        return self.f_prime(self.inputs) * gradients


class Tanh(Activation):
    """
    Tanh activation
    """

    def __init__(self) -> None:
        super().__init__(tanh, tanh_prime)
